#include <stdio.h>
#include "qemu/osdep.h"
#include "qemu/thread.h"
#include <sys/epoll.h>

#include "hw/i386/x86.h"
#include "hw/pci/msi.h"

#ifndef INT_SVC_H
#define INT_SVC_H

#define EPOLL_MAX_EVENTS 10
// MAX 0xff, change types in protocol when increasing limit
#define INT_BUCKET_SIZE 256

// MAX 0xff, change types in protocol when increasing limit
#define SOURCE_BUCKET_SIZE 0x10

#define SOURCE_GSI 0x00
// #define SOURCE_I8259 0x01
// #define SOURCE_IOAPIC 0x02
#define SOURCE_HW_EDU 0x03

#define INT_SVC_VERSION 0x00

typedef struct int_svc_options
{
    char *addr;
    int port;
} IntSvcOptions;

/**
 * @brief Protocol struct in TLV
 * Ops
 * 0x00 ping
 * 0x01 version
 * 0x02 get interception status
 * 0x03 get interception stats
 * 0x04 toggle timer interrupt interception
 * 0x05 toggle interruption by source
 * 0x06 get interception list
 * 0x07 pass interception by source and intno
 */
typedef struct int_svc_protocol_tlv
{
    uint8_t type;
    uint32_t length;
    uint8_t *value;
} IntSvcProtocolTlv;

typedef struct int_svc_interrupt_requests
{
    void *opaque;
    int is_set; // 1 if the int is set, reset if we replay the signal
    int count;  // Count of intercepted requests, reset if we replay the signal
} IntSvcInterruptRequests;

typedef struct int_svc_interceptions
{
    IntSvcInterruptRequests *requests; // Array of requests
    int is_router;                     // 1 if the int was captured on IC like i8259, ioapic, 0 if captured on hardwarez
    void *opaque;
    int is_set; // Only used when is_router is 0, reset if we replay the signal
    int count;  // Count of intercepted requests, reset if we replay the signal
} IntSvcInterceptions;

typedef struct int_svc_interception_response
{
    int source;
    int intno;
    int count;
} IntSvcInterceptionResponse;

typedef struct int_svc_client_context
{
    // May add more fields
    int fd;
    int is_tlv_set;
    int buf_size;
    int buf_limit;
    uint8_t *buf;
    IntSvcProtocolTlv tlv;
} IntSvcClientContext;

/**
 * @brief Parse options from qemu command line and initialize the service
 */
void int_svc_parse_options(const char *options);

/**
 * @brief Management server epoll thread
 */
void *int_svc_service_run(void *arg);

/**
 * @brief Log interrupts in other threads (basically in iothread gsi_handler, i8259, ioapic)
 */
void int_svc_log_int_with_source(int source, int intno, int level);

/**
 * @brief Log interrupts in CPU thread
 */
void int_svc_log_int_cpu(int intvec);

/**
 * @brief Check whether the timer interrupt should be intercepted
 *
 * @return int 1 if intercepted, 0 if not
 */
int int_svc_timer_int_intercept(void);

/**
 * @brief Check whether the interrupt should be intercepted
 *
 * @return int 1 if intercepted, 0 if not
 */
int int_svc_intercept_other_int(int source, void *opaque, int irq, int level);

void int_svc_service_clean_up(void);

#endif